"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const ts_simple_ast_1 = require("ts-simple-ast");
const ts = __importStar(require("typescript"));
function moveDeclarationNamed(declarationName, file, project, targetFile) {
    let found;
    file.forEachDescendant((child, stop) => {
        if (child.getKindName().endsWith('Declaration') && (child.getName && child.getName()) === declarationName) {
            found = child;
            stop();
        }
    });
    if (!found) {
        throw new Error(`declaration named ${declarationName} not found in file ${file.getFilePath()}`);
    }
    return moveDeclaration(found, project, targetFile);
}
exports.moveDeclarationNamed = moveDeclarationNamed;
/**
 * Moves a top level declaration to another file. See all todo comments along the code to  get idea of current
 * state / issue. The resulting project state should not have errors. We do two important things:
 *
 * 1) fix all imports pointing to target declaration in all files of the project so they point now to the new
 *    location (target file)
 * 2) all nodes referenced inside the target declaration must be now imported in the target file (so the
 *    declaration can still see them)
 *
 * **Very WIP dont use it in production yet!**
 *
 * @param c the declaration node to move
 * @param project project in which to do the refactor
 * @param targetFile the target file, inside the project where to move the class
 */
function moveDeclaration(c, project, targetFile) {
    const originalFile = c.getSourceFile();
    // write declaration at the beggining of target file making sure it's exported
    let declarationText = c.getText();
    declarationText.trim().startsWith('export') ? declarationText : 'export ' + declarationText;
    declarationText = declarationText + '\n';
    targetFile.insertStatements(0, declarationText);
    // add necessary imports in target file so declaration has its dependencies: 
    addDeclarationDependencyImportsToTargetFile(originalFile, project, targetFile, c);
    let modifiedFiles = {};
    if (c.isExported) {
        // if c is not exported then we dont need to perform fixProjectImports
        modifiedFiles = fixProjectImports(c, project, targetFile, originalFile);
    }
    // TODO: support / test exports with custom names like :  "export $DECLARATION as OtherName ""
    c.remove();
    c = null;
    // now we call organizeImports on all required files, target, original and all the ones where we added imports
    // organize imports in all files which imports decls where modified (clean unused/duplicated)
    targetFile.organizeImports();
    originalFile.organizeImports();
    Object.values(modifiedFiles).forEach(file => {
        file.organizeImports();
    });
    project.saveSync();
}
exports.moveDeclaration = moveDeclaration;
/**
 * For each reference inside target declaration we will need to import it in targetFile.
 *
 * TODO: we only consider imported declaration dependencies. but not dependencies in original file.
 * Two alternatives:
 * we can move them too or we can export them and import them in the targe file. (configurable?)
  */
function addDeclarationDependencyImportsToTargetFile(originalFile, project, targetFile, c) {
    const imports = [];
    const importStructures = [];
    originalFile.forEachDescendant(d => {
        if (d.getKind() === ts.SyntaxKind.ImportDeclaration) {
            const importDecl = d;
            let moduleSpecifierSourceFile;
            // create a importDeclStructure and reassign the moduleSpecifier to the correct path
            const importDeclStructure = importDeclaration2Structure(importDecl);
            importDeclStructure.moduleSpecifier =
                targetFile.getRelativePathAsModuleSpecifierTo(importDecl.getSourceFile().getDirectory()) + '/' +
                    ((moduleSpecifierSourceFile = importDecl.getModuleSpecifierSourceFile()) &&
                        moduleSpecifierSourceFile.getBaseNameWithoutExtension());
            importStructures.push(importDeclStructure);
        }
    });
    // get the index of the last import so we add imports at the end of the file (so organize imports will
    // remove them in case of duplicates)
    const currentImports = targetFile.getChildrenOfKind(ts.SyntaxKind.SyntaxList)[0].getChildrenOfKind(ts.SyntaxKind.ImportDeclaration);
    let insertPos = 0;
    if (currentImports && currentImports.length) {
        insertPos = currentImports[currentImports.length - 1].getChildIndex() + 1;
    }
    importStructures.forEach(is => {
        targetFile.insertImportDeclarations(insertPos, importStructures);
    });
    // The same for all types referenced inside declaration
    c.forEachDescendant(d => {
        if (ts_simple_ast_1.TypeGuards.isIdentifier(d) && d.getText() !== c.getName()) {
            let definitionNodes;
            if ((definitionNodes = d.getDefinitionNodes()).find(d => d.getSourceFile().getFilePath() === c.getSourceFile().getFilePath() &&
                !!d /*ExportableNode*/.hasExportKeyword)) // this way we differentiate between (for ex) ClassDeclaration and PropertyDeclaration. We want only exportable things
             {
                targetFile.addImportDeclaration({
                    namedImports: [d.getText()],
                    moduleSpecifier: targetFile.getRelativePathAsModuleSpecifierTo(d.getSourceFile())
                });
            }
        }
    });
}
/**
 * Search where target declaration is being imported, in the entire project, and update the imports
 * so it points to the new location (targetFile)
 * @param c
 * @param project
 * @param targetFile
 */
function fixProjectImports(c, project, targetFile, originalFile) {
    let parent;
    const importsReferencing = getAllReferencesToNodeSymbol(c).map(r => r.getNode())
        .filter(n => (parent = n.getParent()) && parent.getKind() === ts.SyntaxKind.ImportSpecifier)
        .map(n => n.getAncestors().find(p => p.getKind() == ts.SyntaxKind.ImportDeclaration));
    // TODO:  ts.SyntaxKind.ImportEqualsDeclaration - contemplate that kind of node - i think it's import a = require('b')
    // TODO: contemplate import {x as y} from Y 
    // TODO: contempoate import * as pp from Y
    // TODO: contemplate internal classes (not classes declared inside functions for ex)
    let oneImportTargetDeclaration;
    const importsFiles = {};
    importsReferencing.forEach(importDecl => {
        if (targetFile.getFilePath().toString() === importDecl.getSourceFile().getFilePath().toString()) {
            // targetFile imports c - we need to remove that import because now declaration will be here
            targetFile.getImportDeclarations().forEach(i => {
                const importSpecifierToC = i.getNamedImports().find(ni => ni.getText() === c.getName());
                if (importSpecifierToC) {
                    importSpecifierToC.remove();
                }
                if (i.getNamedImports().length === 0) {
                    i.remove();
                }
            });
        }
        else {
            importDecl.setModuleSpecifier(importDecl.getSourceFile().
                getRelativePathAsModuleSpecifierTo(targetFile.getDirectory()) + '/' + targetFile.getBaseNameWithoutExtension());
            oneImportTargetDeclaration = importDecl;
        }
        importsFiles[importDecl.getSourceFile().getFilePath()] = importDecl.getSourceFile();
    });
    // Also we want to import declaration in original file (cause other parts of the file could still being
    // referencing it.). We take advantage of these import decls here and do it here... 
    const importDeclStructure = importDeclaration2Structure(oneImportTargetDeclaration);
    importDeclStructure.moduleSpecifier = originalFile.getRelativePathAsModuleSpecifierTo(targetFile.getDirectory()) +
        '/' + targetFile.getBaseNameWithoutExtension();
    originalFile.addImportDeclaration(importDeclStructure);
    importsFiles[originalFile.getFilePath()] = originalFile;
    return importsFiles;
}
// utilities
/**
 * I don't know why, but this will return more results than just c.getReferencingNodes() and is what I need
 * now.
 * TODO: investigate why and contribute
 */
function getAllReferencesToNodeSymbol(c) {
    const results = [];
    for (const referencedSymbol of c.findReferences()) {
        for (const reference of referencedSymbol.getReferences()) {
            results.push(reference);
        }
    }
    return results;
}
/**
 * @param importDecl creates a ImportDeclarationStructure from given ImportDeclaration
 */
function importDeclaration2Structure(importDecl) {
    let defImport;
    let namedImport;
    let namespaceImport;
    let moduleSpecifier;
    return {
        namedImports: (namedImport = importDecl.getNamedImports()) && namedImport.map(ni => ni.getText()),
        defaultImport: (defImport = importDecl.getDefaultImport()) && defImport.getText(),
        namespaceImport: (namespaceImport = importDecl.getNamespaceImport()) && namespaceImport.getText(),
        moduleSpecifier: importDecl.getModuleSpecifier().getText()
    };
}
//# sourceMappingURL=moveDeclaration.js.map